@@ -330,6 +330,10 @@ class OEMHandler(generic.OEMHandler):
     def is_fpc(self):
         """True if the target is a Lenovo nextscale fan power controller
         """
+        if self.has_imm or self.has_xcc:
+            return None
+        if self._fpc_variant is not None:
+            return self._fpc_variant
         fpc_ids = ((19046, 32, 1063), (20301, 32, 462))
         smm_id = (19046, 32, 1180)
         currid = (self.oemid['manufacturer_id'], self.oemid['device_id'],
@@ -381,29 +385,29 @@ class OEMHandler(generic.OEMHandler):
                 yield inv
 
     def get_sensor_data(self):
-        if self.is_fpc:
-            for name in nextscale.get_sensor_names(self._fpc_variant):
-                yield nextscale.get_sensor_reading(name, self.ipmicmd,
-                                                   self._fpc_variant)
-        elif self.has_imm:
+        if self.has_imm:
             for name in self.immhandler.get_oem_sensor_names(self.ipmicmd):
                 yield self.immhandler.get_oem_sensor_reading(name,
                                                              self.ipmicmd)
+        elif self.is_fpc:
+            for name in nextscale.get_sensor_names(self._fpc_variant):
+                yield nextscale.get_sensor_reading(name, self.ipmicmd,
+                                                   self._fpc_variant)
 
     def get_sensor_descriptions(self):
-        if self.is_fpc:
-            return nextscale.get_sensor_descriptions(self._fpc_variant)
-        elif self.has_imm:
+        if self.has_imm:
             return self.immhandler.get_oem_sensor_descriptions(self.ipmicmd)
+        elif self.is_fpc:
+            return nextscale.get_sensor_descriptions(self._fpc_variant)
         return ()
 
     def get_sensor_reading(self, sensorname):
-        if self.is_fpc:
-            return nextscale.get_sensor_reading(sensorname, self.ipmicmd,
-                                                self._fpc_variant)
-        elif self.has_imm:
+        if self.has_imm:
             return self.immhandler.get_oem_sensor_reading(sensorname,
                                                           self.ipmicmd)
+        elif self.is_fpc:
+            return nextscale.get_sensor_reading(sensorname, self.ipmicmd,
+                                                self._fpc_variant)
         return ()
 
     def get_inventory_of_component(self, component):
@@ -607,6 +611,10 @@ class OEMHandler(generic.OEMHandler):
                                               self._fpc_variant)
         return super(OEMHandler, self).get_oem_firmware(bmcver)
 
+    def get_diagnostic_data(self, savefile, progress):
+        if self.has_xcc:
+            return self.immhandler.get_diagnostic_data(savefile, progress)
+
     def get_oem_capping_enabled(self):
         if self.has_tsm:
             rsp = self.ipmicmd.xraw_command(netfn=0x3a, command=0x1b,
@@ -652,6 +660,8 @@ class OEMHandler(generic.OEMHandler):
                                                 data=(4, i))
                 name += rsp['data'][:]
             return name.rstrip('\x00')
+        elif self.is_fpc:
+            return self.smmhandler.get_domain()
 
     def set_oem_domain_name(self, name):
         if self.has_tsm:
@@ -669,6 +679,22 @@ class OEMHandler(generic.OEMHandler):
 
             self._restart_dns()
             return
+        elif self.is_fpc:
+            self.smmhandler.set_domain(name)
+
+    def set_hostname(self, hostname):
+        if self.has_xcc:
+            return self.immhandler.set_hostname(hostname)
+        elif self.is_fpc:
+            return self.smmhandler.set_hostname(hostname)
+        return super(OEMHandler, self).set_hostname(hostname)
+
+    def get_hostname(self):
+        if self.has_xcc:
+            return self.immhandler.get_hostname()
+        elif self.is_fpc:
+            return self.smmhandler.get_hostname()
+        return super(OEMHandler, self).get_hostname()
 
     """ Gets a remote console launcher for a Lenovo ThinkServer.
 
@@ -896,16 +922,26 @@ class OEMHandler(generic.OEMHandler):
                                                        progress=progress,
                                                        bank=bank)
 
-    def get_system_configuration(self):
+    def get_description(self):
+        if self.has_xcc:
+            return self.immhandler.get_description()
+        return super(OEMHandler, self).get_description()
+
+    def get_system_configuration(self, hideadvanced):
         if self.has_imm or self.has_xcc:
-            return self.immhandler.get_system_configuration()
-        return super(OEMHandler, self).get_system_configuration()
+            return self.immhandler.get_system_configuration(hideadvanced)
+        return super(OEMHandler, self).get_system_configuration(hideadvanced)
 
     def set_system_configuration(self, changeset):
         if self.has_imm or self.has_xcc:
             return self.immhandler.set_system_configuration(changeset)
         return super(OEMHandler, self).set_system_configuration(changeset)
 
+    def clear_system_configuration(self):
+        if self.has_xcc:
+            return self.immhandler.clear_system_configuration()
+        return super(OEMHandler, self).clear_system_configuration()
+
     def detach_remote_media(self):
         if self.has_imm:
             self.immhandler.detach_remote_media()
@@ -915,11 +951,16 @@ class OEMHandler(generic.OEMHandler):
             self.ipmicmd.xraw_command(netfn=0x32, command=0x9f, data=(8, 11))
 
     def upload_media(self, filename, progress):
-        if self.has_xcc:
+        if self.has_xcc or self.has_imm:
             return self.immhandler.upload_media(filename, progress)
         return super(OEMHandler, self).upload_media(filename, progress)
 
     def list_media(self):
-        if self.has_xcc:
+        if self.has_xcc or self.has_imm:
             return self.immhandler.list_media()
         return super(OEMHandler, self).list_media()
+
+    def get_health(self, summary):
+        if self.has_xcc:
+            return self.immhandler.get_health(summary)
+        return super(OEMHandler, self).get_health(summary)
