@@ -14,8 +14,10 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import base64
 from datetime import datetime
 import errno
+import fnmatch
 import json
 import os.path
 import pyghmi.constants as pygconst
@@ -37,7 +39,15 @@ import weakref
 
 
 numregex = re.compile('([0-9]+)')
-
+funtypes = {
+    0: 'RAID Controller',
+    1: 'Ethernet',
+    2: 'Fibre Channel',
+    3: 'Infiniband',
+    4: 'GPU',
+    10: 'NVMe Controller',
+    12: 'Fabric Controller',
+}
 
 def naturalize_string(key):
     """Analyzes string in a human way to enable natural sort
@@ -86,6 +96,13 @@ class IMMClient(object):
     ADP_LABEL = 'adapter.connectorLabel'
     ADP_SLOTNO = 'adapter.slotNo'
     ADP_OOB = 'adapter.oobSupported'
+    ADP_PARTNUM = 'vpd.partNo'
+    ADP_SERIALNO = 'vpd.serialNo'
+    ADP_VENID = 'generic.vendorId'
+    ADP_SUBVENID = 'generic.subVendor'
+    ADP_DEVID = 'generic.devId'
+    ADP_SUBDEVID = 'generic.subDevId'
+    ADP_FRU = 'vpd.cardSKU'
     BUSNO = 'generic.busNo'
     PORTS = 'network.pPorts'
     DEVNO = 'generic.devNo'
@@ -175,11 +192,13 @@ class IMMClient(object):
             self.fwovintage = util._monotonic_time()
         for key in list(changeset):
             if key not in self.fwo:
+                found = False
                 for rkey in self.fwo:
-                    if rkey.lower() == key.lower():
+                    if fnmatch.fnmatch(rkey.lower(), key.lower()):
                         changeset[rkey] = changeset[key]
-                        del changeset[key]
-                        break
+                        found = True
+                if found:
+                    del changeset[key]
                 else:
                     raise pygexc.InvalidParameterValue(
                         '{0} not a known setting'.format(key))
@@ -447,6 +466,14 @@ class IMMClient(object):
                             except ValueError:
                                 pass
                         yield ('{0} {1}'.format(aname, fname), bdata)
+        for disk in self.disk_inventory():
+            yield disk
+        self.weblogout()
+
+    def disk_inventory(self, mode=0):
+        if mode==1:
+            # Bypass IMM hardware inventory for now
+            return
         storagedata = self.get_cached_data('lenovo_cached_storage')
         if not storagedata:
             if self.wc:
@@ -472,7 +499,6 @@ class IMMClient(object):
                     bdata['version'] = diskent['storage.firmwares'][0][
                         'versionStr']
                     yield (diskname, bdata)
-        self.weblogout()
 
     def get_hw_inventory(self):
         hwmap = self.hardware_inventory_map()
@@ -533,7 +559,8 @@ class IMMClient(object):
         hwmap = {}
         enclosureuuid = self.get_property('/v2/ibmc/smm/chassis/uuid')
         if enclosureuuid:
-            bay = self.get_property('/v2/cmm/sp/7')
+            bay = hex(int(self.get_property('/v2/cmm/sp/7'))).replace(
+                '0x', '')
             serial = self.get_property('/v2/ibmc/smm/chassis/sn')
             model = self.get_property('/v2/ibmc/smm/chassis/mtm')
             hwmap['Enclosure'] = {
@@ -542,6 +569,8 @@ class IMMClient(object):
                 'Model': fixup_str(model),
                 'Serial': fixup_str(serial),
             }
+        for disk in self.disk_inventory(mode=1): # hardware mode
+            hwmap[disk[0]] = disk[1]
         adapterdata = self.get_cached_data('lenovo_cached_adapters')
         if not adapterdata:
             if self.updating:
@@ -557,14 +586,6 @@ class IMMClient(object):
             anames = {}
             for adata in adapterdata['items']:
                 skipadapter = False
-                if not adata[self.ADP_OOB]:
-                    continue
-                aname = adata[self.ADP_NAME]
-                if aname in anames:
-                    anames[aname] += 1
-                    aname = '{0} {1}'.format(aname, anames[aname])
-                else:
-                    anames[aname] = 1
                 clabel = adata[self.ADP_LABEL]
                 if clabel == 'Unknown':
                     continue
@@ -574,17 +595,44 @@ class IMMClient(object):
                         clabel = 'ML2 (Slot {0})'.format(aslot)
                     else:
                         clabel = 'Slot {0}'.format(aslot)
-                bdata = {'location': clabel}
+                aname = adata[self.ADP_NAME]
+                bdata = {'location': clabel, 'name': aname}
+                if aname in anames:
+                    anames[aname] += 1
+                    aname = '{0} {1}'.format(aname, anames[aname])
+                else:
+                    anames[aname] = 1
                 for fundata in adata[self.ADP_FUN]:
                     bdata['pcislot'] = '{0:02x}:{1:02x}'.format(
                         fundata[self.BUSNO], fundata[self.DEVNO])
-                    serialdata = fundata.get('vpd.serialNo', None)
+                    serialdata = fundata.get(self.ADP_SERIALNO, None)
                     if (serialdata and serialdata != 'N/A' and
                             '---' not in serialdata):
                         bdata['serial'] = serialdata
-                    partnum = fundata.get('vpd.partNo', None)
+                    partnum = fundata.get(self.ADP_PARTNUM, None)
                     if partnum and partnum != 'N/A':
-                        bdata['partnumber'] = partnum
+                        bdata['Part Number'] = partnum
+                    cardtype = funtypes.get(fundata.get('funType', None),
+                                            None)
+                    if cardtype is not None:
+                        bdata['Type'] = cardtype
+                    venid = fundata.get(self.ADP_VENID, None)
+                    if venid is not None:
+                        bdata['PCI Vendor ID'] = '{0:04x}'.format(venid)
+                    devid = fundata.get(self.ADP_DEVID, None)
+                    if devid is not None:
+                        bdata['PCI Device ID'] = '{0:04x}'.format(devid)
+                    venid = fundata.get(self.ADP_SUBVENID, None)
+                    if venid is not None:
+                        bdata['PCI Subsystem Vendor ID'] = '{0:04x}'.format(
+                            venid)
+                    devid = fundata.get(self.ADP_SUBDEVID, None)
+                    if devid is not None:
+                        bdata['PCI Subsystem Device ID'] = '{0:04x}'.format(
+                            devid)
+                    fruno = fundata.get(self.ADP_FRU, None)
+                    if fruno is not None:
+                        bdata['FRU Number'] = fruno
                     if self.PORTS in fundata:
                         for portinfo in fundata[self.PORTS]:
                             for lp in portinfo['logicalPorts']:
@@ -671,6 +719,13 @@ class XCCClient(IMMClient):
     ADP_LABEL = 'connectorLabel'
     ADP_SLOTNO = 'slotNo'
     ADP_OOB = 'oobSupported'
+    ADP_PARTNUM = 'vpd_partNo'
+    ADP_SERIALNO = 'vpd_serialNo'
+    ADP_VENID = 'generic_vendorId'
+    ADP_SUBVENID = 'generic_subVendor'
+    ADP_DEVID = 'generic_devId'
+    ADP_SUBDEVID = 'generic_subDevId'
+    ADP_FRU = 'vpd_cardSKU'
     BUSNO = 'generic_busNo'
     PORTS = 'network_pPorts'
     DEVNO = 'generic_devNo'
@@ -684,6 +739,18 @@ class XCCClient(IMMClient):
         dsc = dsc[0]
         return {'height': int(dsc['u-height']), 'slot': int(dsc['slot'])}
 
+    def clear_system_configuration(self):
+        res = self.wc.grab_json_response(
+                '/redfish/v1/Systems/1/Bios/Actions/Bios.ResetBios',
+                {'Action': 'Bios.ResetBios'},
+                headers={'Authorization': 'Basic ' + base64.b64encode(
+                    self.username + ':' + self.password),
+                         'Content-Type': 'application/json'})
+        if not res:
+            raise Exception(
+                'Unexpected response to clear configuration: {0}'.format(
+                    self.wc.lastjsonerror))
+
     def get_webclient(self, login=True):
         cv = self.ipmicmd.certverify
         wc = webclient.SecureHTTPConnection(self.imm, 443, verifycallback=cv)
@@ -729,15 +796,78 @@ class XCCClient(IMMClient):
         rsp = self.wc.grab_json_response(
             '/api/function/raid_conf?params=raidlink_GetStatus')
         if rsp['items'][0]['status'] != 2:
-            raise pygexc.TemporaryError('Storage configuration unavilable in '
+            raise pygexc.TemporaryError('Storage configuration unavailable in '
                                         'current state (try boot to setup or '
                                         'an OS)')
         if not cfgspec:
             return True
         for pool in cfgspec.arrays:
             self._parse_storage_cfgspec(pool)
+        self.weblogout()
         return True
 
+    def get_diagnostic_data(self, savefile, progress=None):
+        self.wc.grab_json_response('/api/providers/ffdc',
+                                   {'Generate_FFDC': 1})
+        percent = 0
+        while percent != 100:
+            ipmisession.Session.pause(3)
+            result = self.wc.grab_json_response('/api/providers/ffdc',
+                                                {'Generate_FFDC_status': 1})
+            self._refresh_token()
+            if progress:
+                progress({'phase': 'initializing', 'progress': float(percent)})
+            percent = result['progress']
+        while 'FileName' not in result:
+            result = self.wc.grab_json_response('/api/providers/ffdc',
+                                                {'Generate_FFDC_status': 1})
+        url = '/ffdc/{0}'.format(result['FileName'])
+        fd = webclient.FileDownloader(self.wc, url, savefile)
+        fd.start()
+        while fd.isAlive():
+            fd.join(1)
+            if progress and self.wc.get_download_progress():
+                progress({'phase': 'download',
+                          'progress': 100 * self.wc.get_download_progress()})
+            self._refresh_token()
+        if progress:
+            progress({'phase': 'complete'})
+
+    def disk_inventory(self, mode=0):
+        # mode 0 is firmware, 1 is hardware
+        storagedata = self.get_cached_data('lenovo_cached_storage')
+        if not storagedata:
+            if self.wc:
+                storagedata = self.wc.grab_json_response(
+                    '/api/function/raid_alldevices?params=storage_GetAllDisks')
+                if storagedata:
+                    self.datacache['lenovo_cached_storage'] = (
+                        storagedata, util._monotonic_time())
+        if storagedata and 'items' in storagedata:
+            for adp in storagedata['items']:
+                for diskent in adp.get('disks', ()):
+                    if mode==0:
+                        yield self.get_disk_firmware(diskent)
+                    elif mode==1:
+                        yield self.get_disk_hardware(diskent)
+
+    def get_disk_hardware(self, diskent):
+        bdata = {}
+        diskname = 'Disk {0}'.format(diskent['slotNo'])
+        bdata['Model'] = diskent['productName'].rstrip()
+        bdata['Serial Number'] = diskent['serialNo'].rstrip()
+        bdata['FRU Number'] = diskent['fruPartNo'].rstrip()
+        bdata['Description'] = diskent['type'].rstrip()
+        return (diskname, bdata)
+
+    def get_disk_firmware(self, diskent):
+        bdata = {}
+        diskname = 'Disk {0}'.format(diskent['slotNo'])
+        bdata['model'] = diskent[
+            'productName'].rstrip()
+        bdata['version'] = diskent['fwVersion']
+        return (diskname, bdata)
+
     def _parse_array_spec(self, arrayspec):
         controller = None
         if arrayspec.disks:
@@ -839,14 +969,15 @@ class XCCClient(IMMClient):
     def clear_storage_arrays(self):
         rsp = self.wc.grab_json_response(
             '/api/function', {'raidlink_ClearRaidConf': '1'})
+        self.weblogout()
         if rsp['return'] != 0:
             raise Exception('Unexpected return to clear config: ' + repr(rsp))
 
     def remove_storage_configuration(self, cfgspec):
-        realcfg = self.get_storage_configuration()
+        realcfg = self.get_storage_configuration(False)
         for pool in cfgspec.arrays:
             for volume in pool.volumes:
-                vid = str(volume.id[1])
+                vid = '{0},{1}'.format(volume.id[1], volume.id[0])
                 rsp = self.wc.grab_json_response(
                     '/api/function', {'raidlink_RemoveVolumeAsync': vid})
                 if rsp['return'] != 0:
@@ -855,9 +986,10 @@ class XCCClient(IMMClient):
                 self._wait_storage_async()
         for disk in cfgspec.disks:
             self._make_available(disk, realcfg)
+        self.weblogout()
 
     def apply_storage_configuration(self, cfgspec):
-        realcfg = self.get_storage_configuration()
+        realcfg = self.get_storage_configuration(False)
         for disk in cfgspec.disks:
             if disk.status.lower() == 'jbod':
                 self._make_jbod(disk, realcfg)
@@ -869,6 +1001,7 @@ class XCCClient(IMMClient):
         for pool in cfgspec.arrays:
             if pool.disks:
                 self._create_array(pool)
+        self.weblogout()
 
     def _create_array(self, pool):
         params = self._parse_array_spec(pool)
@@ -886,7 +1019,7 @@ class XCCClient(IMMClient):
             if vol.name is None:
                 # need to iterate while there exists a volume of that name
                 if currvolnames is None:
-                    currcfg = self.get_storage_configuration()
+                    currcfg = self.get_storage_configuration(False)
                     currvolnames = set([])
                     for pool in currcfg.arrays:
                         for volume in pool.volumes:
@@ -951,12 +1084,12 @@ class XCCClient(IMMClient):
             drives.append(drive)
         return controller
 
-    def get_storage_configuration(self):
+    def get_storage_configuration(self, logout=True):
         rsp = self.wc.grab_json_response(
             '/api/function/raid_alldevices?params=storage_GetAllDevices')
         standalonedisks = []
         pools = []
-        for item in rsp['items']:
+        for item in rsp.get('items', []):
             for cinfo in item['controllerInfo']:
                 cid = cinfo['id']
                 for pool in cinfo['pools']:
@@ -986,13 +1119,15 @@ class XCCClient(IMMClient):
                         disks=disks, raid=pool['rdlvlstr'], volumes=volumes,
                         id=(cid, pool['id']), hotspares=spares,
                         capacity=totalsize, available_capacity=freesize))
-                for disk in cinfo['unconfiguredDisks']:
+                for disk in cinfo.get('unconfiguredDisks', ()):
                     # can be unused, global hot spare, or JBOD
                     standalonedisks.append(
                         storage.Disk(
                             name=disk['name'], description=disk['type'],
                             id=(cid, disk['id']), status=disk['RAIDState'],
                             serial=disk['serialNo'], fru=disk['fruPartNo']))
+        if logout:
+            self.weblogout()
         return storage.ConfigSpec(disks=standalonedisks, arrays=pools)
 
     def attach_remote_media(self, url, user, password):
