@@ -14,11 +14,11 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import pyghmi.constants as const
+
 IPMI_BMC_ADDRESS = 0x20
 IPMI_SEND_MESSAGE_CMD = 0x34
 
-import pyghmi.constants as const
-
 
 payload_types = {
     'ipmi': 0x0,
@@ -1285,7 +1285,7 @@ sensor_type_offsets = {
             'deassertion_severity': const.Health.Ok,
         },
     },
-    0x21:  {  # slot/connector
+    0x21: {  # slot/connector
         0x0: {
             'desc': 'Fault',
             'severity': const.Health.Critical,
@@ -1877,3 +1877,11 @@ ipmi_completion_codes = {
     0xff: "Unspecified",
     0xffff: "Timeout",  # not ipmi, but used internally
 }
+
+priv_levels = {
+     'callback': 1,
+     'user': 2,
+     'operator': 3,
+     'admin': 4,
+     'oem': 5,
+}
\ No newline at end of file
