@@ -27,7 +27,9 @@ import socket
 import struct
 import threading
 
-from Crypto.Cipher import AES
+
+from cryptography.hazmat.backends import default_backend
+from cryptography.hazmat.primitives.ciphers import algorithms, Cipher, modes
 
 import pyghmi.exceptions as exc
 from pyghmi.ipmi.private import constants
@@ -116,18 +118,23 @@ def define_worker():
                 for w in iowaiters:
                     w[1].set()
                 iowaiters = []
-                for d in directediowaiters:
+                for d in list(directediowaiters):
                     # these are the existing waiters that didn't get
                     # satisfied last graball, allow them to set a new
                     # deadline if they still have time waiting, or
                     # if they have expired, wake them now to let them
                     # process their timeout
-                    for w in directediowaiters[d]:
-                        ltimeout = w[0] - _monotonic_time()
+                    rightnow = _monotonic_time()
+                    for idx, w in reversed(list(
+                            enumerate(directediowaiters[d]))):
+                        ltimeout = w[0] - rightnow
                         if ltimeout < 0:
                             w[1].set()  # time is up, wake the caller
+                            del directediowaiters[d][idx]
                         elif ltimeout < timeout:
                             timeout = ltimeout
+                    if not directediowaiters[d]:
+                        del directediowaiters[d]
                 while ioqueue:
                     workitem = ioqueue.popleft()
                     # order: function, args, list to append to , event to set
@@ -246,18 +253,18 @@ def _aespad(data):
     """ipmi demands a certain pad scheme,
     per table 13-20 AES-CBC encrypted payload fields.
     """
-    newdata = list(data)
     currlen = len(data) + 1  # need to count the pad length field as well
     neededpad = currlen % 16
     if neededpad:  # if it happens to be zero, hurray, but otherwise invert the
         # sense of the padding
         neededpad = 16 - neededpad
     padval = 1
+    pad = bytearray(neededpad)
     while padval <= neededpad:
-        newdata.append(padval)
+        pad[padval - 1] = padval
         padval += 1
-    newdata.append(neededpad)
-    return newdata
+    pad.append(neededpad)
+    return pad
 
 
 def _checksum(*data):  # Two's complement over the data
@@ -302,6 +309,10 @@ class Session(object):
     # can do something like reassign our threading and select modules
     socketchecking = None
 
+    # Maintain single Cryptography backend for all IPMI sessions (seems to be
+    # thread-safe)
+    _crypto_backend = default_backend()
+
     @classmethod
     def _cleanup(cls):
         for sesskey in list(cls.bmc_handlers):
@@ -400,10 +411,10 @@ class Session(object):
         with util.protect(KEEPALIVE_SESSIONS):
             sess = cls.keepalive_sessions.get(session, None)
             if sess is not None and 'timeout' in sess:
-                if sess['timeout'] < _monotonic_time():
+                if sess['timeout'] < _monotonic_time() - 15:
                     # session would have timed out by now, don't use it
                     return False
-        return True
+        return not session.broken
 
     def __new__(cls,
                 bmc,
@@ -411,7 +422,8 @@ class Session(object):
                 password,
                 port=623,
                 kg=None,
-                onlogon=None):
+                onlogon=None,
+                privlevel=4):
         trueself = None
         forbidsock = []
         for res in socket.getaddrinfo(bmc, port, 0, socket.SOCK_DGRAM):
@@ -458,11 +470,12 @@ class Session(object):
                  password,
                  port=623,
                  kg=None,
-                 onlogon=None):
+                 onlogon=None,
+                 privlevel=4):
         if hasattr(self, 'initialized'):
             # new found an existing session, do not corrupt it
             if onlogon is None:
-                while self.logging:
+                while self.logging and not self.broken:
                     Session.wait_for_rsp()
             else:
                 if self.logging:
@@ -471,7 +484,9 @@ class Session(object):
                     self.iterwaiters.append(onlogon)
             return
         self.broken = False
-        self.privlevel = 4
+        self.socket = None
+        self.logged = 0
+        self.privlevel = privlevel
         self.maxtimeout = 3  # be aggressive about giving up on initial packet
         self.incommand = False
         self.nameonly = 16  # default to name only lookups in RAKP exchange
@@ -507,17 +522,17 @@ class Session(object):
             self.kg = self.password
         self.port = port
         if onlogon is None:
-            self.async = False
+            self.async_ = False
             self.logonwaiters = [self._sync_login]
         else:
-            self.async = True
+            self.async_ = True
             self.logonwaiters = [onlogon]
         if self.__class__.socketchecking is None:
             self.__class__.socketchecking = threading.Lock()
         with self.socketchecking:
             self.socket = self._assignsocket(forbiddensockets=self.forbidsock)
         self.login()
-        if not self.async:
+        if not self.async_:
             while self.logging:
                 Session.wait_for_rsp()
         if self.broken:
@@ -526,41 +541,24 @@ class Session(object):
     def _mark_broken(self, error=None):
         # since our connection has failed retries
         # deregister our keepalive facility
+        self.lastpayload = None
         with util.protect(KEEPALIVE_SESSIONS):
             Session.keepalive_sessions.pop(self, None)
         with util.protect(WAITING_SESSIONS):
             Session.waiting_sessions.pop(self, None)
+        try:
+            del Session.initting_sessions[(self.bmc, self.userid,
+                                           self.password, self.port,
+                                           self.kgo)]
+        except KeyError:
+            pass
+        self.logout()
         self.logging = False
         self.errormsg = error
-        if self.logged:
-            self.logged = 0  # mark session as busted
-            self.logging = False
-            if self._customkeepalives:
-                for ka in list(self._customkeepalives):
-                    # Be thorough and notify parties through their custom
-                    # keepalives.  In practice, this *should* be the same, but
-                    # if a code somehow makes duplicate SOL handlers,
-                    # this would notify all the handlers rather than just the
-                    # last one to take ownership
-                    self._customkeepalives[ka][1](
-                        {'error': 'Session Disconnected'})
-            self._customkeepalives = None
-            if not self.broken:
-                self.socketpool[self.socket] -= 1
-                self.broken = True
-                # since this session is broken, remove it from the handler list
-                # This allows constructor to create a new, functional object to
-                # replace this one
-                myport = self.socket.getsockname()[1]
-                for sockaddr in self.allsockaddrs:
-                    if (sockaddr in Session.bmc_handlers and
-                            myport in Session.bmc_handlers[sockaddr]):
-                        del Session.bmc_handlers[sockaddr][myport]
-                        if Session.bmc_handlers[sockaddr] == {}:
-                            del Session.bmc_handlers[sockaddr]
-        elif not self.broken:
+        if not self.broken:
             self.broken = True
-            self.socketpool[self.socket] -= 1
+            if self.socket:
+                self.socketpool[self.socket] -= 1
 
     def onlogon(self, parameter):
         if 'error' in parameter:
@@ -614,16 +612,16 @@ class Session(object):
         """This function generate message for bridge request. It is a
         part of ipmi payload.
         """
-        head = [constants.IPMI_BMC_ADDRESS,
-                constants.netfn_codes['application'] << 2]
+        head = bytearray((constants.IPMI_BMC_ADDRESS,
+                          constants.netfn_codes['application'] << 2))
         check_sum = _checksum(*head)
         # NOTE(fengqian): according IPMI Figure 14-11, rqSWID is set to 81h
-        boday = [0x81, self.seqlun, constants.IPMI_SEND_MESSAGE_CMD,
-                 0x40 | channel]
+        boday = bytearray((0x81, self.seqlun, constants.IPMI_SEND_MESSAGE_CMD,
+                           0x40 | channel))
         # NOTE(fengqian): Track request
         self._add_request_entry((constants.netfn_codes['application'] + 1,
                                  self.seqlun, constants.IPMI_SEND_MESSAGE_CMD))
-        return head + [check_sum] + boday
+        return head + bytearray((check_sum,)) + boday
 
     def _add_request_entry(self, entry=()):
         """This function record the request with netfn, sequence number and
@@ -677,12 +675,12 @@ class Session(object):
         # figure 13-4, first two bytes are rsaddr and
         # netfn, for non-bridge request, rsaddr is always 0x20 since we are
         # addressing BMC while rsaddr is specified forbridge request
-        header = [rsaddr, netfn << 2]
+        header = bytearray((rsaddr, netfn << 2))
 
-        reqbody = [rqaddr, self.seqlun, command] + list(data)
-        headsum = _checksum(*header)
-        bodysum = _checksum(*reqbody)
-        payload = header + [headsum] + reqbody + [bodysum]
+        reqbody = bytearray((rqaddr, self.seqlun, command)) + data
+        headsum = bytearray((_checksum(*header),))
+        bodysum = bytearray((_checksum(*reqbody),))
+        payload = header + headsum + reqbody + bodysum
         if bridge_request:
             payload = bridge_msg + payload
             # NOTE(fengqian): For bridge request, another check sum is needed.
@@ -704,6 +702,8 @@ class Session(object):
             stillin = self.incommand - _monotonic_time()
             if stillin > 0:
                 return stillin
+            else:
+                self.lastpayload = None
         return 0
 
     def _getmaxtimeout(self):
@@ -785,11 +785,13 @@ class Session(object):
         if retry is None:
             retry = not self.servermode
         if self.servermode:
-            data = [code] + list(data)
+            data = bytearray((code,)) + bytearray(data)
             if netfn is None:
                 netfn = self.clientnetfn
             if command is None:
                 command = self.clientcommand
+        else:
+            data = bytearray(data)
         ipmipayload = self._make_ipmi_payload(netfn, command, bridge_request,
                                               data)
         payload_type = constants.payload_types['ipmi']
@@ -821,10 +823,12 @@ class Session(object):
             payload_type = self.last_payload_type
         if not payload:
             payload = self.lastpayload
-        message = [0x6, 0, 0xff, 0x07]  # constant RMCP header for IPMI
+        message = bytearray(b'\x06\x00\xff\x07')  # constant IPMI RMCP header
         if retry:
             self.lastpayload = payload
             self.last_payload_type = payload_type
+        if not isinstance(payload, bytearray):
+            payload = bytearray(payload)
         message.append(self.authtype)
         baretype = payload_type
         if self.integrityalgo:
@@ -839,10 +843,10 @@ class Session(object):
             elif baretype not in constants.payload_types.values():
                 raise NotImplementedError(
                     "Unrecognized payload type %d" % baretype)
-            message += struct.unpack("!4B", struct.pack("<I", self.sessionid))
-        message += struct.unpack("!4B", struct.pack("<I", self.sequencenumber))
+            message += struct.pack("<I", self.sessionid)
+        message += struct.pack("<I", self.sequencenumber)
         if self.ipmiversion == 1.5:
-            message += struct.unpack("!4B", struct.pack("<I", self.sessionid))
+            message += struct.pack("<I", self.sessionid)
             if not self.authtype == 0:
                 message += self._ipmi15authcode(payload)
             message.append(len(payload))
@@ -866,18 +870,20 @@ class Session(object):
                 message.append(newpsize & 0xff)
                 message.append(newpsize >> 8)
                 iv = os.urandom(16)
-                message += list(struct.unpack("16B", iv))
-                payloadtocrypt = _aespad(payload)
-                crypter = AES.new(self.aeskey, AES.MODE_CBC, iv)
-                crypted = crypter.encrypt(struct.pack("%dB" %
-                                                      len(payloadtocrypt),
-                                                      *payloadtocrypt))
-                crypted = list(struct.unpack("%dB" % len(crypted), crypted))
-                message += crypted
+                message += iv
+                payloadtocrypt = bytes(payload + _aespad(payload))
+                crypter = Cipher(
+                    algorithm=algorithms.AES(self.aeskey),
+                    mode=modes.CBC(iv),
+                    backend=self._crypto_backend
+                )
+                encryptor = crypter.encryptor()
+                message += encryptor.update(payloadtocrypt
+                                            ) + encryptor.finalize()
             else:  # no confidetiality algorithm
                 message.append(psize & 0xff)
                 message.append(psize >> 8)
-                message += list(payload)
+                message += payload
             if self.integrityalgo:  # see table 13-8,
                 # RMCP+ packet format
                 # TODO(jbjohnso): SHA256 which is now
@@ -885,18 +891,15 @@ class Session(object):
                 neededpad = (len(message) - 2) % 4
                 if neededpad:
                     neededpad = 4 - neededpad
-                message += [0xff] * neededpad
+                message += b'\xff' * neededpad
                 message.append(neededpad)
                 message.append(7)  # reserved, 7 is the required value for the
                 # specification followed
-                integdata = message[4:]
-                authcode = hmac.new(self.k1,
-                                    struct.pack("%dB" % len(integdata),
-                                                *integdata),
+                message += hmac.new(self.k1,
+                                    bytes(message[4:]),
                                     hashlib.sha1).digest()[:12]  # SHA1-96
                 # per RFC2404 truncates to 96 bits
-                message += struct.unpack("12B", authcode)
-        self.netpacket = struct.pack("!%dB" % len(message), *message)
+        self.netpacket = message
         # advance idle timer since we don't need keepalive while sending
         # packets out naturally
         with util.protect(KEEPALIVE_SESSIONS):
@@ -918,19 +921,14 @@ class Session(object):
         if padneeded < 0:
             raise exc.IpmiException("Password is too long for ipmi 1.5")
         password += '\x00' * padneeded
-        passdata = struct.unpack("16B", password)
         if checkremotecode:
-            seqbytes = struct.unpack("!4B",
-                                     struct.pack("<I", self.remsequencenumber))
+            seqbytes = struct.pack("<I", self.remsequencenumber)
         else:
-            seqbytes = struct.unpack("!4B",
-                                     struct.pack("<I", self.sequencenumber))
-        sessdata = struct.unpack("!4B", struct.pack("<I", self.sessionid))
-        bodydata = passdata + sessdata + tuple(payload) + seqbytes + passdata
-        dgst = hashlib.md5(
-            struct.pack("%dB" % len(bodydata), *bodydata)).digest()
-        hashdata = struct.unpack("!%dB" % len(dgst), dgst)
-        return hashdata
+            seqbytes = struct.pack("<I", self.sequencenumber)
+        sessdata = struct.pack("<I", self.sessionid)
+        bodydata = password + sessdata + payload + seqbytes + password
+        dgst = hashlib.md5(bodydata).digest()
+        return dgst
 
     def _got_channel_auth_cap(self, response):
         if 'error' in response:
@@ -1271,8 +1269,8 @@ class Session(object):
             # things off ignore the second reply since we have one
             # satisfactory answer
         if data[4] in (0, 2):  # This is an ipmi 1.5 paylod
-            remsequencenumber = struct.unpack('<I', bytes(data[5:9]))[0]
-            remsessid = struct.unpack("<I", bytes(data[9:13]))[0]
+            remsequencenumber = struct.unpack('<I', data[5:9])[0]
+            remsessid = struct.unpack("<I", data[9:13])[0]
             if (remsequencenumber == 0 and remsessid == 0 and
                     qent[2] in Session.bmc_handlers):
                 # So a new ipmi client happens to get a previously seen and
@@ -1295,20 +1293,15 @@ class Session(object):
 
             if remsessid != self.sessionid:
                 return -1  # does not match our session id, drop it
-            # now we need a mutable representation of the packet, rather than
-            # copying pieces of the packet over and over
-            rsp = list(struct.unpack("!%dB" % len(data), bytes(data)))
             authcode = False
             if data[4] == 2:  # we have authcode in this ipmi 1.5 packet
                 authcode = data[13:29]
-                del rsp[13:29]
+                del data[13:29]
                 # this is why we needed a mutable representation
-            payload = list(rsp[14:14 + rsp[13]])
+            payload = data[14:14 + data[13]]
             if authcode:
                 expectedauthcode = self._ipmi15authcode(payload,
                                                         checkremotecode=True)
-                expectedauthcode = struct.pack("%dB" % len(expectedauthcode),
-                                               *expectedauthcode)
                 if expectedauthcode != authcode:
                     return
             self._parse_ipmi_payload(payload)
@@ -1358,13 +1351,13 @@ class Session(object):
             if self.k1 is None:  # we are in no shape to process a packet now
                 return
             expectedauthcode = hmac.new(
-                self.k1, bytes(data[4:-12]), hashlib.sha1).digest()[:12]
+                self.k1, data[4:-12], hashlib.sha1).digest()[:12]
             if authcode != expectedauthcode:
                 return  # BMC failed to assure integrity to us, drop it
-            sid = struct.unpack("<I", bytes(data[6:10]))[0]
+            sid = struct.unpack("<I", data[6:10])[0]
             if sid != self.localsid:  # session id mismatch, drop it
                 return
-            remseqnumber = struct.unpack("<I", bytes(data[10:14]))[0]
+            remseqnumber = struct.unpack("<I", data[10:14])[0]
             if (hasattr(self, 'remseqnumber') and
                 (remseqnumber < self.remseqnumber) and
                     (self.remseqnumber != 0xffffffff)):
@@ -1374,13 +1367,16 @@ class Session(object):
             payload = data[16:16 + psize]
             if encrypted:
                 iv = data[16:32]
-                decrypter = AES.new(self.aeskey, AES.MODE_CBC, bytes(iv))
-                decrypted = decrypter.decrypt(
-                    struct.pack("%dB" % len(payload[16:]),
-                                *payload[16:]))
-                payload = struct.unpack("%dB" % len(decrypted), decrypted)
+                crypter = Cipher(
+                    algorithm=algorithms.AES(self.aeskey),
+                    mode=modes.CBC(bytes(iv)),
+                    backend=self._crypto_backend
+                )
+                decryptor = crypter.decryptor()
+                payload = bytearray(decryptor.update(bytes(payload[16:])
+                                                     ) + decryptor.finalize())
                 padsize = payload[-1] + 1
-                payload = list(payload[:-padsize])
+                payload = payload[:-padsize]
             if ptype == 0:
                 self._parse_ipmi_payload(payload)
             elif ptype == 1:  # There should be no other option
@@ -1570,6 +1566,15 @@ class Session(object):
         # For now, skip the checksums since we are in LAN only,
         # TODO(jbjohnso): if implementing other channels, add checksum checks
         # here
+        if len(payload) < 7:
+            # This cannot possibly be a valid IPMI packet.  Note this is after
+            # the integrity checks, so this must be a buggy BMC packet
+            # One example was a BMC that if receiving an SOL deactivate
+            # from another party would emit what looks to be an attempt
+            # at SOL deactivation payload, but with the wrong payload type
+            # since we can't do anything remotely sane with such a packet,
+            # drop it and carry about our business.
+            return
         if self.servermode:
             self.seqlun = payload[4]
             self.clientaddr = payload[3]
@@ -1621,8 +1626,7 @@ class Session(object):
         # doing anything, though it shouldn't matter
         self.lastpayload = None
         self.last_payload_type = None
-        response = {}
-        response['netfn'] = payload[1] >> 2
+        response = {'netfn': payload[1] >> 2}
         # ^^ remove header of rsaddr/netfn/lun/checksum/rq/seq/lun
         del payload[0:5]
         # remove the trailing checksum
@@ -1656,6 +1660,7 @@ class Session(object):
             self._mark_broken()
             return
         elif self.sessioncontext == 'FAILED':
+            self.lastpayload = None
             self.nowait = False
             return
         if self.sessioncontext == 'OPENSESSION':
@@ -1747,11 +1752,34 @@ class Session(object):
                                             struct.pack("I", self.sessionid)),
                          retry=False)
         # stop trying for a keepalive,
+        self.lastpayload = None
         with util.protect(KEEPALIVE_SESSIONS):
             Session.keepalive_sessions.pop(self, None)
         self.logged = 0
         self.logging = False
+        if self._customkeepalives:
+            for ka in list(self._customkeepalives):
+                # Be thorough and notify parties through their custom
+                # keepalives.  In practice, this *should* be the same, but
+                # if a code somehow makes duplicate SOL handlers,
+                # this would notify all the handlers rather than just the
+                # last one to take ownership
+                self._customkeepalives[ka][1](
+                    {'error': 'Session Disconnected'})
         self._customkeepalives = None
+        if not self.broken:
+            self.socketpool[self.socket] -= 1
+            self.broken = True
+            # since this session is broken, remove it from the handler list
+            # This allows constructor to create a new, functional object to
+            # replace this one
+            myport = self.socket.getsockname()[1]
+            for sockaddr in self.allsockaddrs:
+                if (sockaddr in Session.bmc_handlers and
+                        myport in Session.bmc_handlers[sockaddr]):
+                    del Session.bmc_handlers[sockaddr][myport]
+                    if Session.bmc_handlers[sockaddr] == {}:
+                        del Session.bmc_handlers[sockaddr]
         self.nowait = False
         self.socketpool[self.socket] -= 1
         return {'success': True}
